!*********************************************************************
!> \brief Grid module
!
!> This module contains the variables and the subroutines related to 
!> the computational grid
!*********************************************************************
MODULE geometry

  IMPLICIT NONE

  !> Location of the centers of the control volume of the domain
  REAL*8, ALLOCATABLE :: z_comp(:) 

  !> Location of the boundaries of the control volumes of the domain
  REAL*8, ALLOCATABLE :: z_stag(:)

  !> Radius at the boundaries of the control volumes of the domain
  REAL*8, ALLOCATABLE :: radius_stag(:)

  !> Eccentricity at the boundaries of the control volumes of the domain
  REAL*8, ALLOCATABLE :: eccen_stag(:)

  REAL*8 :: z0           !< Left (bottom) of the physical domain
  REAL*8 :: zN           !< Right (top) of the physical domain
  REAL*8 :: zeta_exit    !< Right (top) of the physical domain
  REAL*8 :: dz           !< Control volumes size
  REAL*8 :: pi

  REAL*8 :: radius       !< Effective radius
  REAL*8 :: f_eccen_a    !< Eccentricity factor a (rate between ellipse perimeter and 2*pi*sqrt(Ra*Rb))
  REAL*8 :: f_eccen_b    !< Eccentricity factor b (sqrt(2*Ra*Ra*Rb*Rb/(Ra*Ra+Rb*Rb))/sqrt(Ra*Rb))

  INTEGER :: n_radius_models

  CHARACTER (LEN=30), DIMENSION(20) :: available_radius_models

  !> geometry model\n
  !> - 'Cylinder'                   => constant radius in depth and circular cross section
  !> - 'Dyke'                       => constant radius in depth and circular elliptical section
  !> - 'Depth_dependent_1'          => linearly enlarging conduit (radius_min = dimensions at bottom; radius_max = dimensions at vent)
  !> - 'Depth_dependent_2'          => cylindrical lower portion and linearly enlarging upper zone
  !> - 'Depth_dependent_3'          => two cylindrical portions connected by a linearly enlarging transitional zone
  !> .

  CHARACTER*30 :: radius_model

  REAL*8 :: radius_fixed     !< Fixed value for radius (cylinder)
  REAL*8 :: radius_a         !< Major semi-axis (dyke)
  REAL*8 :: radius_b         !< Minor semi-axis (dyke)
  REAL*8 :: radius_min       !< Minimum radius (depth_dependent_1, depth_dependent_2, depth_dependent_3) 
  REAL*8 :: radius_max       !< Maximum radius (depth_dependent_1, depth_dependent_2, depth_dependent_3) 
  REAL*8 :: radius_z         !< Depth for conduit enlargement (depth_dependent_2, depth_dependent_3) 
  REAL*8 :: radius_z_sig     !< Length of transitional zone (depth_dependent_3)
  INTEGER :: comp_cells      !< Number of control volumes in the computational domain
  INTEGER :: comp_interfaces !< Number of interfaces (comp_cells+1)

CONTAINS


  SUBROUTINE initialize_geometry_models

    IMPLICIT NONE

    n_radius_models = 5

    available_radius_models(1) = 'Cylinder'
    available_radius_models(2) = 'Dyke'
    available_radius_models(3) = 'Depth_dependent_1' 
    available_radius_models(4) = 'Depth_dependent_2' 
    available_radius_models(5) = 'Depth_dependent_3' 

  END SUBROUTINE initialize_geometry_models

  !*********************************************************************
  !> \brief Finite volume grid initialization
  !
  !> This subroutine initialize the grids for the finite volume solver.
  !> \date 16/08/2011
  !*********************************************************************

  SUBROUTINE init_grid

    IMPLICIT none

    INTEGER j      !> loop counter

    pi = 4.D0 * DATAN(1.D0)

    comp_interfaces = comp_cells+1

    ALLOCATE( z_comp(comp_cells) )
    ALLOCATE( z_stag(comp_interfaces) )
    ALLOCATE( radius_stag(comp_interfaces) )
    ALLOCATE( eccen_stag(comp_interfaces) )

    dz = ( zN - z0 ) / comp_cells

    z_stag(1) = z0
    z_comp(1) = z0 + 0.5 * dz
    
    DO j=1,comp_cells

       z_stag(j+1) = MIN( z_stag(j) + dz , zN )

       z_comp(j) = 0.5 * ( z_stag(j) + z_stag(j+1) )

    END DO

    SELECT CASE ( radius_model )
 
    CASE DEFAULT

       radius_stag(1:comp_interfaces) = radius_fixed

       eccen_stag(1:comp_interfaces) = 0.0D0

    CASE ('Cylinder' )

       radius_stag(1:comp_interfaces) = radius_fixed
       
       eccen_stag(1:comp_interfaces) = 0.0D0

    CASE ('Dyke' )

       radius_stag(1:comp_interfaces) = SQRT( radius_a * radius_b )
       
       eccen_stag(1:comp_interfaces) = SQRT( 1.0D0 - ( MIN(radius_b, radius_a) ** 2.0D0 ) / ( MAX(radius_b, radius_a) ** 2.0D0 ) )

    CASE ( 'Depth_dependent_1' )

       DO j=1,comp_interfaces
          
          radius_stag(j) = radius_min + (radius_max - radius_min) * z_stag(j) / (zN - z0)
          
       END DO

       eccen_stag(1:comp_interfaces) =0.0D0

    CASE ( 'Depth_dependent_2' )

       DO j=1,comp_interfaces
          
          IF( z_stag(j) < (zN - z0 - radius_z) ) THEN
             
             radius_stag(j) = radius_min
             
          ELSE

             radius_stag(j) = radius_min + (radius_max - radius_min) *                    &
                  ( z_stag(j) - zN + z0 + radius_z ) / ( radius_z )
             
          END IF
          
       END DO
 
       eccen_stag(1:comp_interfaces) = 0.0D0
 
    CASE ( 'Depth_dependent_3' )
      
       DO j=1,comp_interfaces
          
          IF( z_stag(j) < (zN - z0 - radius_z) ) THEN
             
             radius_stag(j) = radius_min
             
          ELSEIF( z_stag(j) < (zN - z0 - radius_z + radius_z_sig) ) THEN
             
             radius_stag(j) = radius_min + (radius_max - radius_min) * &
                  ( z_stag(j) - zN + z0 + radius_z  ) / ( radius_z_sig )
             
          ELSE
             
             radius_stag(j) = radius_max
             
          END IF
          
       END DO

       eccen_stag(1:comp_interfaces) = 0.0D0

    END SELECT

  END SUBROUTINE init_grid

  SUBROUTINE update_radius(zeta)

    IMPLICIT NONE

    REAL*8, INTENT(IN) :: zeta
    REAL*8 :: coeff_interp, eccen_aux1, eccen_aux2
    INTEGER :: j , z_idx

    z_idx = 1

    DO j=1,comp_interfaces-1

       IF ( z_stag(j) < zeta ) z_idx = j

    END DO    

    ! zeta is between z_stag(j) and z_stag(j+1)
    
    coeff_interp = ( zeta - z_stag(z_idx) ) / ( z_stag(z_idx+1) - z_stag(z_idx) )  

    radius = coeff_interp * radius_stag(z_idx+1) + ( 1.D0 - coeff_interp ) *        &
         radius_stag(z_idx)

    eccen_aux1 = coeff_interp * eccen_stag(z_idx+1) + ( 1.D0 - coeff_interp ) *          &
         eccen_stag(z_idx)

    eccen_aux2 = (1.D0 - eccen_aux1**2.D0)**(0.5D0)

    f_eccen_a =  ( 3.D0 * (1.D0 + eccen_aux2 ) - ((3.D0 + eccen_aux2) *  &
		              (1.D0 + 3.D0*eccen_aux2 ))**(0.5D0) ) / (2.D0 * eccen_aux2**(0.5D0) )

    f_eccen_b =  SQRT( (2.D0 * eccen_aux2 ) / (1 + eccen_aux2**2.D0) )

  END SUBROUTINE update_radius

END MODULE geometry
